import sys
import Image
import ImageFilter

class MyPixelAccess:
    """
    An improved version of the PIL's standart class, PixelAccess. Along with the
    image itself stores its size and mode.
    """
    def __init__(self, im):
        self.pix = im.load()
        self.size = im.size
        self.mode = im.mode
    def fit(self, x, y):
        return (0 <= x < im.size[0]) and (0 <= y < im.size[1])
    def __getitem__(self, coords):
        return self.pix[coords]
    def __setitem__(self, coords, value):
        self.pix[coords] = value
    def save(self):
        im = Image.new(self.mode, self.size)
        for y in range(self.size[1]):
            for x in range(self.size[0]):
                im.putpixel((x, y), self.pix[x, y])
        return im

def neighbours4(x, y):
    return [(x-1,y), (x, y-1), (x+1, y), (x, y+1)]

def neighbours8(x, y):
    return [(x-1, y-1), (x, y-1), (x+1, y-1), (x+1, y), \
            (x+1, y+1), (x, y+1), (x-1, y+1), (x-1, y)]

getNeighbours = neighbours4

def addColours(c1, c2):
    return (c1[0] + c2[0], c1[1] + c2[1], c1[2] + c2[2])

def difference(c1, c2):
    s1 = c1[0] + c1[1] + c1[2]
    s2 = c2[0] + c2[1] + c2[2]
    result = abs(s1 - s2)
    return result

def divideColour(c1, n):
    return (c1[0] / n, c1[1] / n, c1[2] / n)

def smartEdge(im, limit):
    pix = MyPixelAccess(im)
    pixW, pixH = im.size
    limit = limit * 3 # this is because we don't want to calculate the average of the RGB components
    areaMap = MyPixelAccess(Image.new("RGB", (pixW, pixH)))
    areaColour = pix[0, 0]
    areaCount = 0

    def notInArea(x, y):
        return areaMap[x, y] == (0, 0, 0)

    def findArea(x, y):
        stack = [(x, y)]
        area = [(x, y)]
        summedAreaColour = (0, 0, 0)
        while stack != []:
            x1, y1 = stack.pop()
            areaMap[x1, y1] = (1, 1, 1)
            summedAreaColour = addColours(summedAreaColour, pix[x1, y1])
            neighbours = filter (lambda (x, y):
                                  pix.fit(x, y) and \
                                  notInArea(x, y) and \
                                  difference(pix[x,y], pix[x1, y1]) <= limit, getNeighbours(x1, y1))
            stack.extend(neighbours)
            area.extend(neighbours)
        square = len(area)
        averageColour = divideColour (summedAreaColour, square)
        for x, y in area:
            areaMap[x, y] = averageColour

    for y in range(pixH):
        for x in range(pixW):
            if notInArea(x, y):
                areaCount += 1
                areaColour = pix[x, y]
                findArea(x, y)
    print "%d / %d" % (areaCount, pixW * pixH)
    return areaMap



if __name__ == "__main__":
    for infile in sys.argv[1:]:
        im = Image.open(infile)
        print "%s: %s, %s - %s" % (infile, im.format, "%dx%d" % im.size, im.mode)
        im = im.convert("RGB")
        smartEdge(im, 4).save().show()